///////////////////////////////////////////////////////////////////////////////
///
///	\file    CoordTransforms.h
///	\author  Paul Ullrich
///	\version March 14, 2020
///
///	<remarks>
///		Copyright 2000-2014 Paul Ullrich
///
///		This file is distributed as part of the Tempest source code package.
///		Permission is granted to use, copy, modify and distribute this
///		source code and its documentation under the terms of the GNU General
///		Public License.  This software is provided "as is" without express
///		or implied warranty.
///	</remarks>

#ifndef _COORDTRANSFORMS_H_
#define _COORDTRANSFORMS_H_

///////////////////////////////////////////////////////////////////////////////

#include <cmath>

#include "Defines.h"
#include "Exception.h"

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Convert radians to degrees.
///	</summary>
inline double RadToDeg(
	double dRad
) {
	return (dRad * 180.0 / M_PI);
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Convert degrees to radians.
///	</summary>
inline double DegToRad(
	double dDeg
) {
	return (dDeg * M_PI / 180.0);
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Translate a longitude value to the range [0,360)
///	</summary>
inline double LonDegToStandardRange(
	double dLonDeg
) {
	dLonDeg = (dLonDeg - 360.0 * floor(dLonDeg / 360.0));
	if ((dLonDeg < 0.0) || (dLonDeg >= 360.0)) {
		return 0.0;
	}
	return dLonDeg;
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Translate a longitude value to the range [0,360)
///	</summary>
inline double LonRadToStandardRange(
	double dLonRad
) {
	dLonRad = (dLonRad - (2.0 * M_PI) * floor(dLonRad / (2.0 * M_PI)));
	if ((dLonRad < 0.0) || (dLonRad >= (2.0 * M_PI))) {
		return 0.0;
	}
	return dLonRad;
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate 3D Cartesian coordinates from latitude and longitude,
///		in radians.
///	</summary>
inline void RLLtoXYZ_Rad(
	double dLonRad,
	double dLatRad,
	double & dX,
	double & dY,
	double & dZ
) {
	_ASSERT(fabs(dLatRad) <= 0.5 * M_PI + HighTolerance);

	dX = cos(dLonRad) * cos(dLatRad);
	dY = sin(dLonRad) * cos(dLatRad);
	dZ = sin(dLatRad);
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate 3D Cartesian coordinates from latitude and longitude,
///		in degrees.
///	</summary>
inline void RLLtoXYZ_Deg(
	double dLonDeg,
	double dLatDeg,
	double & dX,
	double & dY,
	double & dZ
) {
	return RLLtoXYZ_Rad(
		DegToRad(dLonDeg),
		DegToRad(dLatDeg),
		dX, dY, dZ);
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate latitude and longitude from normalized 3D Cartesian
///		coordinates, in degrees.
///	</summary>
inline void XYZtoRLL_Deg(
	double dX,
	double dY,
	double dZ,
	double & dLonDeg,
	double & dLatDeg
) {
	double dMag2 = dX * dX + dY * dY + dZ * dZ;

	if (fabs(dMag2 - 1.0) >= 0.01) {
		_EXCEPTION4("Grid point has non-unit magnitude: "
			"(%1.15e, %1.15e, %1.15e) (magnitude %1.15e)",
			dX, dY, dZ, fabs(dX * dX + dY * dY + dZ * dZ));

	}

	double dMag = sqrt(dMag2);

	dX /= dMag;
	dY /= dMag;
	dZ /= dMag;

	if (fabs(dZ) < 1.0 - ReferenceTolerance) {
		dLonDeg = RadToDeg(atan2(dY, dX));
		dLatDeg = RadToDeg(asin(dZ));

		if (dLonDeg < 0.0) {
			dLonDeg += 360.0;
		}

	} else if (dZ > 0.0) {
		dLonDeg = 0.0;
		dLatDeg = 90.0;

	} else {
		dLonDeg = 0.0;
		dLatDeg = -90.0;
	}
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate latitude and longitude from normalized 3D Cartesian
///		coordinates, in radians.
///	</summary>
inline void XYZtoRLL_Rad(
	double dX,
	double dY,
	double dZ,
	double & dLonRad,
	double & dLatRad
) {
	double dMag2 = dX * dX + dY * dY + dZ * dZ;

	if (fabs(dMag2 - 1.0) >= 0.01) {
		_EXCEPTION4("Grid point has non-unit magnitude: "
			"(%1.15e, %1.15e, %1.15e) (magnitude %1.15e)",
			dX, dY, dZ, fabs(dX * dX + dY * dY + dZ * dZ));

	}

	double dMag = sqrt(dMag2);

	dX /= dMag;
	dY /= dMag;
	dZ /= dMag;

	if (fabs(dZ) < 1.0 - ReferenceTolerance) {
		dLonRad = atan2(dY, dX);
		dLatRad = asin(dZ);

		if (dLonRad < 0.0) {
			dLonRad += 2.0 * M_PI;
		}

	} else if (dZ > 0.0) {
		dLonRad = 0.0;
		dLatRad = 0.5 * M_PI;

	} else {
		dLonRad = 0.0;
		dLatRad = -0.5 * M_PI;
	}

}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate an average longitude from two given longitudes (in radians).
///	</summary>
inline double AverageLongitude_Rad(
	double dLonRad1,
	double dLonRad2
) {
	double dDeltaLonRad;
	if (dLonRad2 > dLonRad1) {
		dDeltaLonRad = fmod(dLonRad2 - dLonRad1, 2.0 * M_PI);
		if (dDeltaLonRad > M_PI) {
			dDeltaLonRad = dDeltaLonRad - 2.0 * M_PI;
		}
	} else {
		dDeltaLonRad = - fmod(dLonRad1 - dLonRad2, 2.0 * M_PI);
		if (dDeltaLonRad < -M_PI) {
			dDeltaLonRad = dDeltaLonRad + 2.0 * M_PI;
		}
	}

	double dLonRadAvg = dLonRad1 + 0.5 * dDeltaLonRad;

	if ((dLonRadAvg < 0.0) && (dLonRad1 >= 0.0) && (dLonRad2 >= 0.0)) {
		dLonRadAvg += 2.0 * M_PI;
	}
	if ((dLonRadAvg > 2.0 * M_PI) && (dLonRad1 <= 2.0 * M_PI) && (dLonRad2 <= 2.0 * M_PI)) {
		dLonRadAvg -= 2.0 * M_PI;
	}

	return dLonRadAvg;
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate the great circle distance (in radians) between two RLL points.
///	</summary>
inline double GreatCircleDistance_Rad(
	double dLonRad1,
	double dLatRad1,
	double dLonRad2,
	double dLatRad2
) {
	double dR =
		sin(dLatRad1) * sin(dLatRad2)
		+ cos(dLatRad1) * cos(dLatRad2) * cos(dLonRad2 - dLonRad1);

	if (dR >= 1.0) {
		dR = 0.0;
	} else if (dR <= -1.0) {
		dR = M_PI;
	} else {
		dR = acos(dR);
	}
	if (dR != dR) {
		_EXCEPTIONT("NaN value detected");
	}

	return dR;
}

///////////////////////////////////////////////////////////////////////////////

///	<summary>
///		Calculate the great circle distance (in degrees) between two RLL points.
///	</summary>
inline double GreatCircleDistance_Deg(
	double dLonRad1,
	double dLatRad1,
	double dLonRad2,
	double dLatRad2
) {
	return
		RadToDeg(
			GreatCircleDistance_Rad(
				dLonRad1,
				dLatRad1,
				dLonRad2,
				dLatRad2));
}

///////////////////////////////////////////////////////////////////////////////

inline void StereographicProjection(
	double dLonRad0,
	double dLatRad0,
	double dLonRad,
	double dLatRad,
	double & dXs,
	double & dYs
) {
	// Forward projection using equations (1)-(3)
	// http://mathworld.wolfram.com/StereographicProjection.html
	double dK = 2.0 / (1.0 + sin(dLatRad0) * sin(dLatRad) + cos(dLatRad0) * cos(dLatRad) * cos(dLonRad - dLonRad0));
	dXs = dK * cos(dLatRad) * sin(dLonRad - dLonRad0);
	dYs = dK * (cos(dLatRad0) * sin(dLatRad) - sin(dLatRad0) * cos(dLatRad) * cos(dLonRad - dLonRad0));
}

///////////////////////////////////////////////////////////////////////////////

inline void StereographicProjectionInv(
	double dLonRad0,
	double dLatRad0,
	double dXs,
	double dYs,
	double & dLonRad,
	double & dLatRad
) {
	// Forward projection using equations (3)-(5)
	// http://mathworld.wolfram.com/StereographicProjection.html
	double dRho = sqrt(dXs * dXs + dYs * dYs);
	double dC = 2.0 * atan(0.5 * dRho);

	if (dRho < 1.0e-14) {
		dLatRad = dLatRad0;
		dLonRad = dLonRad0;
		return;
	}

	dLatRad = asin(cos(dC) * sin(dLatRad0) + dYs * sin(dC) * cos(dLatRad0) / dRho);
	dLonRad = dLonRad0 + atan2(
			dXs * sin(dC),
			dRho * cos(dLatRad0) * cos(dC) - dYs * sin(dLatRad0) * sin(dC));
}

///////////////////////////////////////////////////////////////////////////////


#endif // _COORDTRANSFORMS_H_

